#!/usr/bin/env python
# -*- coding: utf-8 -*-

# # 1A.data - Décorrélation de variables aléatoires - correction
# 
# On construit des variables corrélées gaussiennes et on cherche à construire des variables décorrélées en utilisant le calcul matriciel. (correction)

# In[1]:


from jyquickhelper import add_notebook_menu
add_notebook_menu()


# ## Création d'un jeu de données

# ### Q1
# 
# La première étape consiste à construire des variables aléatoires normales corrélées dans une matrice $N \times 3$. On cherche à construire cette matrice au format [numpy](http://www.numpy.org/). Le programme suivant est un moyen de construire un tel ensemble à l'aide de combinaisons linéaires. Complétez les lignes contenant des ``....``.

# In[2]:


import random
import numpy as np

def combinaison():
    x = random.gauss(0,1) # génère un nombre aléatoire
    y = random.gauss(0,1) # selon une loi normale
    z = random.gauss(0,1) # de moyenne null et de variance 1
    x2 = x
    y2 = 3*x + y
    z2 = -2*x + y + 0.2*z
    return [x2, y2, z2]
    
li = [ combinaison () for i in range (0,100) ]
mat = np.matrix(li)
mat[:5]


# ### Q2 

# In[3]:


npm = mat
t   = npm.transpose ()
a   = t @ npm 
a  /= npm.shape[0]
a


# ``a`` est la matrice de covariance.

# ### Corrélation de matrices

# ### Q3

# In[4]:


cov = a


# In[5]:


var = np.array([cov[i,i]**(-0.5) for i in range(cov.shape[0])])
var.resize((3,1))
varvar = var @ var.transpose()
varvar


# In[6]:


cor = np.multiply(cov, varvar)
cor


# ### Q4 

# In[7]:


def correlation(npm):
    t   = npm.transpose ()
    a   = t @ npm 
    a  /= npm.shape[0]
    var = np.array([cov[i,i]**(-0.5) for i in range(cov.shape[0])])
    var.resize((3,1))
    varvar = var @ var.transpose()
    return np.multiply(cov, varvar)

correlation(npm)


# ## Calcul de la racine carrée

# ### Q6
# 
# Le module [numpy](http://www.numpy.org/) propose une fonction qui retourne la matrice $P$ et le vecteur des valeurs propres $L$ :
# 
# ```
# L,P = np.linalg.eig(a)
# ```
# 
# Vérifier que $P'P=I$. Est-ce rigoureusement égal à la matrice identité ?

# In[8]:


L,P = np.linalg.eig(a)


# In[9]:


P.transpose() @ P


# C'est presque l'identité aux erreurs de calcul près.

# ### Q7
# 
# ``np.diag(l)`` construit une matrice diagonale à partir d'un vecteur.

# In[10]:


np.diag(L)


# ### Q8
# 
# Ecrire une fonction qui calcule la racine carrée de la matrice $\frac{1}{n}M'M$ (on rappelle que $M$ est la matrice ``npm``). Voir aussi [Racine carrée d'une matrice](https://fr.wikipedia.org/wiki/Racine_carr%C3%A9e_d%27une_matrice).

# In[11]:


def square_root_matrix(M):
    L,P = np.linalg.eig(M)
    L = L ** 0.5
    root = P @ np.diag(L) @ P.transpose()
    return root
    
root = square_root_matrix(cov)
root


# On vérifie qu'on ne s'est pas trompé.

# In[12]:


root @ root - cov


# ## Décorrélation

# In[13]:


np.linalg.inv(cov)


# ### Q9 
# 
# Chaque ligne de la matrice $M$ représente un vecteur de trois variables corrélées. La matrice de covariance est $V=\frac{1}{n}M'M$. Calculer la matrice de covariance de la matrice $N=M V^{-\frac{1}{2}}$ (mathématiquement).
# 

# ### Q10
# 
# Vérifier numériquement.

# ## Simulation de variables corrélées

# ### Q11
# 
# A partir du résultat précédent, proposer une méthode pour simuler un vecteur de variables corrélées selon une matrice de covariance $V$ à partir d'un vecteur de lois normales indépendantes.

# ### Q12
# 
# Proposer une fonction qui crée cet échantillon :

# In[14]:


def simultation (N, cov) :
    # simule un échantillon de variables corrélées
    # N : nombre de variables
    # cov : matrice de covariance
    # ...
    return M


# ### Q13
# 
# Vérifier que votre échantillon a une matrice de corrélations proche de celle choisie pour simuler l'échantillon.

# In[15]: